import datetime
import os
import pickle
import shutil
import subprocess
import sys
import time

import lab as B
import slugify

from .out import out, kv, streams, Section

__all__ = ["generate_root", "WorkingDirectory"]


def generate_root(name):
    """Generate a root path.

    Args:
        name (str): Name of the experiment.

    Returns:
        str: Root path.
    """
    now = time.strftime("%Y-%m-%d_%H-%M-%S")
    return os.path.join("_experiments", f"{now}_{slugify.slugify(name)}")


class Logger:
    """A logger to a file.

    Args:
        path (str): Path to file.
    """

    def __init__(self, path):
        self.path = path

        # Empty file.
        with open(self.path, "w"):
            pass

    def write(self, message):
        """Write to the file.

        Args:
            message (str): Message to write.
        """
        with open(self.path, "a") as f:
            f.write(message)

    def flush(self):
        """Flush the stream."""
        # Nothing to do.


_default_log = object()


class WorkingDirectory:
    """Working directory.

    Args:
        *root (str): Root of working directory. Use different arguments for
            directories.
        observe (bool, optional): Observe the results. For example, do not overwrite
            the existing log and copied script. Defaults to `False.`
        override (bool, optional): Delete working directory if it already
            exists. Defaults to `False`.
        log (str or None, optional): Initialise logger. Set to `None`
            to disable. Defaults to `log.txt`.
        seed (int, optional): Value of random seed. Defaults to `0`.
    """

    def __init__(self, *root, observe=False, override=False, log=_default_log, seed=0):
        self.root = os.path.join(*root)

        # Delete if the root already exists.
        if os.path.exists(self.root) and override:
            out("Experiment directory already exists. Overwriting.")
            shutil.rmtree(self.root)

        # Create root directory.
        os.makedirs(self.root, exist_ok=True)

        # Set default log.
        if observe:
            if log is _default_log:
                log = None
        else:
            if log is _default_log:
                log = "log.txt"

        # Initialise logger.
        if log is not None:
            streams.append(Logger(self.file(log)))

        kv("Root", self.root)
        kv("Call", " ".join(sys.argv))
        kv("Now", datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
        kv("Python", sys.version)

        # Print details about git.
        try:
            # This command will fail if we are not in a git repo.
            subprocess.check_output(["git", "status"])

            sha = subprocess.check_output(["git", "rev-parse", "HEAD"])
            sha = sha.decode("ascii").strip()
            diff_stat = subprocess.check_output(["git", "diff", "--shortstat"])
            clean = diff_stat.decode("ascii").strip() == ""

            with Section("Git"):
                kv("Commit", sha)
                kv("Dirty", "no" if clean else "yes")
        except subprocess.CalledProcessError:
            out("Git not available.")

        # Copy calling script.
        if not observe:
            path = os.path.abspath(sys.argv[0])
            if os.path.exists(path):
                shutil.copy(path, self.file("script.py"))
            else:
                out("Could not save calling script.")

        # Sed and print random seed.
        B.set_random_seed(seed)
        kv("Seed", seed)

    def file(self, *name, exists=False):
        """Get the path of a file.

        Args:
            *name (str): Path to file, relative to the root directory. Use
                different arguments for directories.
            exists (bool): Assert that the file already exists. Defaults to
                `False`.

        Returns:
            str: Path to file.
        """
        path = os.path.join(self.root, *name)

        # Ensure that path exists.
        if exists and not os.path.exists(path):
            raise AssertionError(f'File "{path}" does not exist.')
        elif not exists:
            path_dir = os.path.join(self.root, *name[:-1])
            os.makedirs(path_dir, exist_ok=True)

        return path

    def save(self, obj, *args, **kw_args):
        """Save an object to a file.

        Further takes in arguments and keyword arguments from
        :meth:`.experiment.WorkingDirectory.file`.

        Args:
            obj (object): Object to save.
        """
        with open(self.file(*args, **kw_args), "wb") as f:
            pickle.dump(obj, f)

    def load(self, *args, **kw_args):
        """Load an object from a file.

        Further takes in arguments and keyword arguments from
        :meth:`.experiment.WorkingDirectory.file`.
        """
        with open(self.file(*args, **kw_args), "rb") as f:
            return pickle.load(f)



import lab as B
from plum import Dispatcher
from stheno import Normal

__all__ = ["elbo"]

_dispatch = Dispatcher()


@_dispatch
def elbo(lik, p, q, num_samples=1):
    """Construct the ELBO.

    Args:
        lik (function): Likelihood that taken in one or more samples from the
            approximate posterior.
        p (distribution): Prior.
        q (distribution): Approximate posterior.
        num_samples (int, optional): Number of samples. Defaults to `1`.

    Returns:
        tensor: ELBO.
    """
    samples = q.sample(num_samples)
    log_lik = B.mean(lik(samples))
    log_p = B.mean(p.logpdf(samples))
    log_q = B.mean(q.logpdf(samples))
    return log_lik + log_p - log_q


@_dispatch
def elbo(lik, p: Normal, q: Normal, num_samples=1):
    return B.mean(lik(q.sample(num_samples))) - q.kl(p)



import datetime

import numpy as np
import pandas as pd

from .data import data_path, split_df, resource

__all__ = ["load"]


def _to_float(x):
    try:
        return float(x)
    except ValueError as e:
        if x in {"", "n/a", " "}:
            return np.nan
        else:
            raise e


def load():
    _fetch()

    sites = ["Bra", "Cam", "Chi", "Sot"]
    features = [("DEPTH", "height"), ("WSPD", "speed"), ("ATMP", "temp")]

    days = list(range(1, 1 + 31))
    fn = "{site:s}{day:02d}Jul2013.csv"

    dfs = []
    for site in sites:
        for day in days:
            df = pd.read_csv(data_path("air_temp", fn.format(site=site, day=day)))

            # Convert times and dates.
            index = []
            for i, row in df.iterrows():
                day, month, year = map(int, row["Date"].split("/"))
                hour, minute = map(int, row["Time"].split(":"))
                dt = datetime.datetime(year, month, day, hour, minute)
                index.append((site, dt))

            # Construct desired data frame.
            index = pd.MultiIndex.from_tuples(index, names=["site", "date"])
            df = pd.DataFrame(
                {
                    name_to: list(map(_to_float, df[name_from]))
                    for name_from, name_to in features
                },
                index=index,
            )
            dfs.append(df)

    # Concat and fix order to match the paper.
    df_all = (
        pd.concat(dfs)
        .unstack("site")
        .reindex(
            [
                ("height", "Bra"),
                ("height", "Sot"),
                ("height", "Cam"),
                ("height", "Chi"),
                ("speed", "Bra"),
                ("speed", "Sot"),
                ("speed", "Cam"),
                ("speed", "Chi"),
                ("temp", "Bra"),
                ("temp", "Sot"),
                ("temp", "Cam"),
                ("temp", "Chi"),
            ],
            axis=1,
        )
    )

    # Clear invalid sensor readings.
    df_all[df_all < -1000] = np.nan

    def make_range(lower, upper):
        day = datetime.timedelta(days=1)
        date_start = df_all.index.min()
        return date_start + (lower - 1) * day, date_start + (upper - 1) * day

    # Construct data sets of various sizes.
    ds = []
    for lower, upper in [
        make_range(10, 20),
        make_range(8, 23),
        (df_all.index.min(), df_all.index.max()),
    ]:
        df = df_all[(df_all.index >= lower) & (df_all.index <= upper)]

        test1, train = split_df(df, make_range(12, 16), [("temp", "Cam")])
        test2, train = split_df(train, make_range(14, 18), [("temp", "Chi")])

        test1_ext, _ = split_df(df, make_range(11.5, 16.5), [("temp", "Cam")])
        test2_ext, _ = split_df(train, make_range(13.5, 18.5), [("temp", "Chi")])

        ds.append((df, train, [test1, test2, test1_ext, test2_ext]))

    return ds


def _fetch():
    site_urls = [
        ("Cam", "http://www.cambermet.co.uk/archive/2013/July/CSV/"),
        ("Chi", "http://www.chimet.co.uk/archive/2013/July/CSV/"),
        ("Bra", "http://www.bramblemet.co.uk/archive/2013/July/CSV/"),
        ("Sot", "http://www.sotonmet.co.uk/archive/2013/July/CSV/"),
    ]
    for day in range(1, 32):
        for site, url in site_urls:
            resource(
                target=data_path("air_temp", f"{site}{day:02d}Jul2013.csv"),
                url=f"{url}{site}{day:02d}Jul2013.csv",
            )



import numpy as np
import pytest
from lab import B
from stheno import EQ

from wbml.lmm import _to_tuples, OLMM, LMMPP
from .util import approx


def test_to_tuples():
    x = np.linspace(0, 1, 3)[:, None]
    y = np.random.randn(3, 2)
    y[0, 0] = np.nan
    y[1, 1] = np.nan

    # Check correctness.
    (x1, i1, y1), (x2, i2, y2) = _to_tuples(x, y)
    approx(x1, x[[1, 2]])
    assert i1 == 0
    approx(y1, y[[1, 2], 0])
    approx(x2, x[[0, 2]])
    assert i2 == 1
    approx(y2, y[[0, 2], 1])

    # Test check that any data is extracted.
    y_nan = y.copy()
    y_nan[:] = np.nan
    with pytest.raises(ValueError):
        _to_tuples(x, y_nan)


def test_lmm_missing_data():
    # Setup model.
    kernels = [EQ(), 2 * EQ().stretch(1.5)]
    noise_obs = 0.1
    noises_latent = np.array([0.1, 0.2])
    H = np.random.randn(3, 2)

    # Construct model.
    lmm = LMMPP(kernels, noise_obs, noises_latent, H)

    # Construct data.
    x = np.linspace(0, 3, 5)
    y = lmm.sample(x, latent=False)

    # Throw away random data points and check that the logpdf computes.
    y2 = y.copy()
    y2[0, 0] = np.nan
    y2[2, 2] = np.nan
    y2[4, 1] = np.nan
    assert not np.isnan(lmm.logpdf(x, y2))

    # Throw away an entire time point and check correctness.
    y2 = y.copy()
    y2[1, :] = np.nan
    approx(lmm.logpdf(x[[0, 2, 3, 4]], y[[0, 2, 3, 4]]), lmm.logpdf(x, y2))

    # Check LML after conditioning.
    lmm = lmm.condition(x, y2)
    approx(lmm.logpdf(x[[0, 2, 3, 4]], y[[0, 2, 3, 4]]), lmm.logpdf(x, y2))


def test_compare_lmm_olmm():
    # Setup models.
    kernels = [EQ(), 2 * EQ().stretch(1.5)]
    noise_obs = 0.1
    noises_latent = np.array([0.1, 0.2])
    U, S, _ = B.svd(B.randn(3, 2))
    H = np.dot(U, np.diag(S) ** 0.5)

    # Construct models.
    lmm = LMMPP(kernels, noise_obs, noises_latent, H)
    olmm = OLMM(kernels, noise_obs, noises_latent, H)

    # Construct data.
    x = np.linspace(0, 3, 5)
    y = lmm.sample(x, latent=False)
    x2 = np.linspace(4, 7, 5)
    y2 = lmm.sample(x2, latent=False)

    # Check LML before conditioning.
    approx(lmm.logpdf(x, y), olmm.logpdf(x, y))
    approx(lmm.logpdf(x2, y2), olmm.logpdf(x2, y2))

    # Check LML after conditioning.
    lmm = lmm.condition(x, y)
    olmm = olmm.condition(x, y)
    # Note: `lmm_pp.lml(x, y)` will not equal `olmm.lml(x, y)` due to
    # assumptions in the OLMM, so the follow is not tested.
    # allclose(lmm.logpdf(x, y), olmm.logpdf(x, y))
    approx(lmm.logpdf(x2, y2), olmm.logpdf(x2, y2))

    # Predict.
    preds_pp, means_pp, vars_pp = lmm.marginals(x2)
    preds, means, vars = olmm.marginals(x2)

    # Check predictions per time point.
    for i in range(5):
        approx(means_pp[i], means[i])
        approx(vars_pp[i], vars[i])

    # Check predictions per output.
    for i in range(3):
        approx(preds_pp[i][0], preds[i][0])
        approx(preds_pp[i][1], preds[i][1])
        approx(preds_pp[i][2], preds[i][2])


class TrackedIterator:
    """An iterator that keeps track of whether it has been used.

    Attributes:
        used (bool): Boolean indicating whether it has been usd.

    Args:
        wrap (object): Object to wrap.
    """

    instances = []

    def __init__(self, wrap):
        self.used = False
        self.wrap = wrap
        TrackedIterator.instances.append(self)

    def __iter__(self):
        self.used = True
        return iter(self.wrap)

    @staticmethod
    def reset():
        """Reset all."""
        for instance in TrackedIterator.instances:
            instance.used = False


def test_lmm_olmm_sample():
    # Setup models.
    kernels = [EQ()] * 2
    noise_obs = 0.1
    noises_latent = np.array([0.1, 0.2])
    H = B.randn(3, 2)

    # Construct models.
    lmm = LMMPP(kernels, noise_obs, noises_latent, H)
    olmm = OLMM(kernels, noise_obs, noises_latent, H)

    # Wrap.
    lmm.fs = TrackedIterator(lmm.fs)
    lmm.ys = TrackedIterator(lmm.ys)
    olmm.xs = TrackedIterator(olmm.xs)
    olmm.xs_noisy = TrackedIterator(olmm.xs_noisy)

    # Test latent samples.
    x = B.randn(10)
    assert isinstance(lmm.sample(x, latent=True), B.NPNumeric)
    assert lmm.fs.used, "lmm.fs was not used."
    assert not lmm.ys.used, "lmm.ys was used."

    assert isinstance(olmm.sample(x, latent=True), B.NPNumeric)
    assert olmm.xs.used, "olmm.xs was not used."
    assert not olmm.xs_noisy.used, "olmm.xs_noisy was used."

    # Test observed samples.
    TrackedIterator.reset()

    assert isinstance(lmm.sample(x, latent=False), B.NPNumeric)
    assert not lmm.fs.used, "lmm.fs was used."
    assert lmm.ys.used, "lmm.ys was not used"

    assert isinstance(olmm.sample(x, latent=False), B.NPNumeric)
    assert not olmm.xs.used, "olmm.xs was used"
    assert olmm.xs_noisy.used, "olmm.xs_noisy was not used."



import importlib
import os.path
import shutil

import matplotlib.pyplot as plt
import pytest

from wbml.data import data_path, DependencyError


class _SetDataAside:
    def __init__(self, name):
        self.path = data_path(name)
        self.moved = False

    def __enter__(self):
        if os.path.exists(self.path):
            # Set data aside.
            os.rename(self.path, self.path + "_set_aside")
            self.moved = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.moved:
            # Remove any data created.
            shutil.rmtree(self.path, ignore_errors=True)
            # Put original data back into place.
            os.rename(self.path + "_set_aside", self.path)


def _import_and_execute(name, kw_args, monkeypatch):
    # Kill showing any plots.
    monkeypatch.setattr(plt, "show", lambda: None)

    module = importlib.import_module(f"wbml.data.{name}")
    module.load(**kw_args)

    # Test any other possible methods.
    for method in ["stats_and_vis"]:
        if hasattr(module, method):
            getattr(module, method)()


@pytest.mark.parametrize(
    "name, kw_args",
    [
        ("air_temp", {}),
        ("eeg", {"extended": False}),
        ("eeg", {"extended": True}),
        ("exchange", {}),
        ("crude_oil", {}),
        ("kemar", {}),
        ("jura", {}),
        ("mauna_loa", {"detrend_method": "gp"}),
        ("mauna_loa", {"detrend_method": "linear"}),
        ("miso", {}),
        ("toy_sines", {}),
        ("vix", {}),
    ],
)
def test_import(name, kw_args, monkeypatch):
    with _SetDataAside(name):
        _import_and_execute(name, kw_args, monkeypatch)


@pytest.mark.parametrize("name", ["cmip5", "station", "snp"])
@pytest.mark.xfail()
def test_import_unable_to_fetch(name, monkeypatch):
    _import_and_execute(name, {}, monkeypatch)


@pytest.mark.parametrize("name", ["cmip5", "station", "snp"])
def test_import_fail_unable_to_fetch(name, monkeypatch):
    with _SetDataAside(name):
        with pytest.raises(DependencyError):
            _import_and_execute(name, {}, monkeypatch)



import lab.tensorflow as B
import tensorflow as tf
import wbml.out as out
from stheno.tensorflow import Normal, Diagonal
from varz.tensorflow import Vars, minimise_adam
from wbml.vi import elbo

d = 2
vs = Vars(tf.float32)

# Define a prior and likelihood and generate data.
prior = Normal(Diagonal(B.linspace(1, 2, d)))
lik_noise = Normal(Diagonal(0.1 * B.ones(d)))
z_true = prior.sample()
y = (lik_noise + z_true).sample(100)


# Define likelihood.
def lik(x):
    return B.sum((lik_noise + y).logpdf(x))


# Construct q distribution.
def construct_q(vs):
    q = Normal(
        Diagonal(vs.positive(shape=[d], name="q/var")),
        vs.get(shape=[d, 1], name="q/mean"),
    )
    return q


# Construct objective.
def objective(vs):
    q = construct_q(vs)
    return -elbo(lik, prior, q, num_samples=1)


# Optimise.
objective_compiled = tf.function(objective, autograph=False)
minimise_adam(objective_compiled, vs, trace=True, iters=2000, rate=5e-2)

# Print result.
out.kv("True", z_true)
out.kv("Mean q", construct_q(vs).mean)



import os
import re
from collections import OrderedDict

import netCDF4 as nc
import numpy as np
import pandas as pd
from datetime import datetime, timedelta

from .data import data_path, asserted_dependency

__all__ = ["load"]


def load():
    _fetch()

    # Load observations.
    data = nc.Dataset(data_path("cmip5", "erai_T2_1979-2016_daily.nc"))
    matrix = data["T2"][:].reshape(data["T2"].shape[0], -1)
    # Convert index to `datetime`s. Not sure what the time values represent...
    times = [
        datetime(year=1979, month=1, day=1) + i * timedelta(days=1)
        for i in range(len(data["time"][:]))
    ]
    obs = pd.DataFrame(matrix, index=pd.Index(times, name="time"))

    # Load locations.
    lat = data["latitude"][:]
    lon = data["longitude"][:]
    lat, lon = np.broadcast_arrays(lat[:, None], lon[None, :])
    assert lat.shape == data["T2"].shape[1:3]
    loc = pd.DataFrame({"latitude": lat.flatten(), "longitude": lon.flatten()})

    # Find simulator files.
    sim_files = os.listdir(data_path("cmip5"))
    sim_files = [
        f
        for f in sim_files
        if os.path.splitext(f)[1].lower() == ".nc"
        if f != "erai_T2_1979-2016_daily.nc"
    ]

    # Load sims.
    sims = OrderedDict()
    for sim_file in sim_files:
        data = nc.Dataset(data_path("cmip5", sim_file))
        matrix = data["tas"][:].reshape(data["tas"].shape[0], -1)
        sim_name = re.match(
            r"^cmip5_tas_amip_(.*)_r1i1p1_1979-2008.nc$", sim_file
        ).group(1)
        sims[sim_name] = pd.DataFrame(
            matrix, index=pd.Index(data["time"][:], name="time")
        )

    return loc, obs, sims


def _fetch():
    files = [
        "cmip5_tas_amip_ACCESS1-0_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_ACCESS1-3_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_BNU-ESM_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_CCSM4_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_CMCC-CM_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_CNRM-CM5_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_CSIRO-Mk3-6-0_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_CanAM4_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_EC-EARTH_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_FGOALS-g2_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_FGOALS-s2_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_GFDL-CM3_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_GFDL-HIRAM-C180_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_GFDL-HIRAM-C360_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_HadGEM2-A_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_IPSL-CM5A-LR_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_IPSL-CM5A-MR_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_IPSL-CM5B-LR_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MIROC5_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MPI-ESM-LR_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MPI-ESM-MR_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MRI-AGCM3-2H_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MRI-AGCM3-2S_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_MRI-CGCM3_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_NorESM1-M_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_bcc-csm1-1-m_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_bcc-csm1-1_r1i1p1_1979-2008.nc",
        "cmip5_tas_amip_inmcm4_r1i1p1_1979-2008.nc",
        "erai_T2_1979-2016_daily.nc",
    ]
    for file in files:
        asserted_dependency(target=data_path("cmip5", file))



import warnings

import pytest
from wbml.warning import warn_upmodule


def test_warn_upmodule(monkeypatch):
    orig_warn = warnings.warn

    def mock_warn(*args, **kw_args):
        assert kw_args["stacklevel"] > 2
        orig_warn(*args, **kw_args)

    monkeypatch.setattr(warnings, "warn", mock_warn)

    with pytest.warns(UserWarning, match="Test warning"):
        warn_upmodule("Test warning", category=UserWarning)


def test_warning_formatting():
    pass



import glob
import multiprocessing
import os
import pickle
import warnings

import numpy as np
import pandas as pd

import wbml.out
from .data import data_path, split_df, resource, dependency

__all__ = ["load", "load_full"]

cache_experiment = data_path("eeg", "experiment.pickle")
cache_experiment_extended = data_path("eeg", "experiment_extended.pickle")
cache_full = data_path("eeg", "full.pickle")


def _load(cache, parse):
    if not os.path.exists(cache):
        parse()
    with open(cache, "rb") as f:
        return pickle.load(f)


def load(extended=False):
    """Load the EEG data, which is a particular split of the first trial of subject
    337 in the large training data

    Args:
        extended (bool, optional): Return the train test splits of the first
            train of all patients in the large training data. Defaults to `False`.

    Returns:
        tuple[:class:`pd.DataFrame`] or dict[int, :class:`pd.DataFrame`]:
            All data, training data, and test data of the first trail of one
            patient or all patients.
    """

    _fetch_large()
    if extended:
        return _load(cache_experiment_extended, _parse_experiment)
    else:
        return _load(cache_experiment, _parse_experiment)


def load_full():
    """Load the full EEG data.

    Returns:
        dict: Full data.
    """
    _fetch_full()
    return _load(cache_full, _parse_full)


def _fetch_large():
    """Fetch the large training and test data set."""
    # Large training data:
    resource(
        target=data_path("eeg", "SMNI_CMI_TRAIN.tar.gz"),
        url="https://kdd.ics.uci.edu/databases/eeg/SMNI_CMI_TRAIN.tar.gz",
    )
    dependency(
        target=data_path("eeg", "train"),
        source=data_path("eeg", "SMNI_CMI_TRAIN.tar.gz"),
        commands=[
            "tar xzf SMNI_CMI_TRAIN.tar.gz",
            "mv SMNI_CMI_TRAIN train",
            "find train | grep gz$ | xargs gunzip",
        ],
    )
    # Large test data:
    resource(
        target=data_path("eeg", "SMNI_CMI_TEST.tar.gz"),
        url="https://kdd.ics.uci.edu/databases/eeg/SMNI_CMI_TEST.tar.gz",
    )
    dependency(
        target=data_path("eeg", "test"),
        source=data_path("eeg", "SMNI_CMI_TEST.tar.gz"),
        commands=[
            "tar xzf SMNI_CMI_TEST.tar.gz",
            "mv SMNI_CMI_TEST test",
            "find test | grep gz$ | xargs gunzip",
        ],
    )


def _fetch_full():
    """Fetch the full data set."""
    resource(
        target=data_path("eeg", "eeg_full.tar"),
        url="https://kdd.ics.uci.edu/databases/eeg/eeg_full.tar",
    )
    dependency(
        target=data_path("eeg", "full"),
        source=data_path("eeg", "eeg_full.tar"),
        commands=[
            "mkdir full",
            "tar xf eeg_full.tar -C full",
            "ls full | grep gz$ | xargs -I {} tar xzf full/{} -C full",
            "ls full | grep gz$ | xargs -I {} rm full/{}",
            "find full | grep gz$ | xargs gunzip",
        ],
    )


def _parse_trial(path):
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        # `np.genfromtxt` produces a warning if the file is empty.
        sites = np.genfromtxt(path, usecols=1, dtype=str)
        data = np.genfromtxt(path, usecols=(2, 3))

    # The file can be empty. In that case, just return nothing.
    if len(sites) == 0:
        return None

    # File is not empty. Start parsing the data.
    x, y = data[:, 0], data[:, 1]

    # Assume that all sites have the same inputs. If that is not true, `np.stack` will
    # fail.
    inds = np.where(np.diff(x) < 0)[0] + 1
    sites = [sites[0]] + [sites[i] for i in inds]
    x = x[: inds[0]] / 256.0  # Sampled at 256 Hz.
    y = np.stack(np.split(y, inds), axis=1)

    return {"df": pd.DataFrame(y, index=pd.Index(x, name="time"), columns=sites)}


def _extract_trials(subject_path):
    paths = glob.glob(f"{subject_path}/*.rd.*")

    # Extract trial numbers.
    trial_numbers = map(lambda x: int(x.split(".rd.")[1]), paths)

    # Parse trials. Use half of all available CPUs.
    with multiprocessing.Pool(processes=multiprocessing.cpu_count() // 2) as pool:
        parsed_trials = pool.map(_parse_trial, paths)

    # Return dictionary mapping trial number to parsed result. Make sure to ignore empty
    # files.
    return {k: v for k, v in zip(trial_numbers, parsed_trials) if v is not None}


def _extract_subjects(path):
    paths = glob.glob(path + "/co*")

    # Safe determine the subject types.
    type_map = {"co2a": "2a", "co2c": "2c", "co3a": "3a", "co3c": "3c"}

    # Extract all subjects.
    subjects = {}
    for path in paths:
        subject = {
            "type": type_map[os.path.split(path)[-1][:4]],
            "trials": _extract_trials(path),
        }
        number = int(os.path.split(path)[-1][-7:])
        subjects[number] = subject

    return subjects


def _parse_experiment():
    wbml.out.out("Parsing EEG data. This should be fairly quick.")

    data = _extract_subjects(data_path("eeg", "train"))

    splits = {}

    for n in data.keys():
        # Just use the first trial of all patients.
        trial_numbers = data[n]["trials"].keys()
        trial = data[n]["trials"][min(trial_numbers)]

        # Select a number of fixed labels.
        labels = ["F3", "F4", "F5", "F6"] + ["FZ", "F1", "F2"]
        trial = trial["df"][labels]

        # Split according to paper.
        test, train = split_df(
            trial,
            index_range=(trial.shape[0] - 100, trial.shape[0]),
            columns=["FZ", "F1", "F2"],
            iloc=True,
        )

        splits[n] = (trial, train, test)

    # Save experiment data.
    with open(cache_experiment, "wb") as f:
        pickle.dump(splits[337], f)

    # Save extended experiment data.
    with open(cache_experiment_extended, "wb") as f:
        pickle.dump(splits, f)


def _parse_full():
    wbml.out.out("Parsing full EEG data. This may take a while.")

    data = _extract_subjects(data_path("eeg", "full"))

    with open(cache_full, "wb") as f:
        pickle.dump(data, f)



import lab as B
from numpy.testing import assert_allclose
from plum import Dispatcher

__all__ = ["approx"]

_dispatch = Dispatcher()


def approx(x, y, rtol=1e-7, atol=0):
    return assert_allclose(B.to_numpy(x), B.to_numpy(y), rtol=rtol, atol=atol)



from abc import ABCMeta, abstractmethod

import lab as B

__all__ = [
    "Normalise",
    "Linear",
    "Activation",
    "Net",
    "Recurrent",
    "GRU",
    "Elman",
    "ff",
    "rnn",
]


class Layer(metaclass=ABCMeta):
    """A layer."""

    @abstractmethod
    def initialise(self, input_size, vs):
        """Initialise the weights of the layer.

        Args:
            input_size (int): Size of the input.
            vs (:class:`varz.Vars`): Variable container.
        """

    @abstractmethod
    def __call__(self, x):  # pragma: no cover
        pass

    @abstractmethod
    def num_weights(self, input_size):
        """Get the number of weights used in the layer.

        Args:
            input_size (int): Size of the input.

        Returns:
            int: Number of weights.
        """

    @property
    @abstractmethod
    def width(self):
        """Width of the layer."""


class Normalise(Layer):
    """Normalisation layer.

    Args:
        epsilon (scalar): Small value to add to the standard deviation before
            dividing. Defaults to `B.epsilon`.
    """

    def __init__(self, epsilon=B.epsilon):
        self._width = None
        self.epsilon = epsilon

    def initialise(self, input_size, vs):
        self._width = input_size

    def __call__(self, x):
        mean = B.mean(x, axis=2)[:, :, None]
        std = B.std(x, axis=2)[:, :, None]
        return (x - mean) / (std + self.epsilon)

    def num_weights(self, input_size):
        self._width = input_size
        return 0

    @property
    def width(self):
        return self._width


class Linear(Layer):
    """Linear layer.

    Args:
        width (int): Width of the layer.
    """

    def __init__(self, width):
        self._width = width
        self.A = None
        self.b = None

    def initialise(self, input_size, vs):
        self.A = vs.get(shape=[input_size, self.width])
        self.b = vs.get(shape=[1, self.width])

    def __call__(self, x):
        return B.matmul(x, self.A) + self.b

    def num_weights(self, input_size):
        return input_size * self.width + self.width

    @property
    def width(self):
        return self._width


class Activation(Layer):
    """Activation layer.

    Args:
        nonlinearity (function, optional): Nonlinearity. Defaults to `B.relu`.
    """

    def __init__(self, nonlinearity=B.relu):
        self._width = None
        self.nonlinearity = nonlinearity

    def initialise(self, input_size, vs):
        self._width = input_size

    def __call__(self, x):
        return self.nonlinearity(x)

    def num_weights(self, input_size):
        self._width = input_size
        return 0

    @property
    def width(self):
        return self._width


class Net(Layer):
    """A composition of layers.

    Args:
        layers (:class:`.net.Layer`): List of the layers to compose.
    """

    def __init__(self, *layers):
        self.layers = layers

    def initialise(self, input_size, vs):
        # Initialise layers.
        for layer in self.layers:
            layer.initialise(input_size, vs)
            input_size = layer.width

    def __call__(self, x):
        x_rank = B.rank(x)
        if x_rank == 1:
            x = x[None, :, None]
        elif x_rank == 2:
            x = x[None, :, :]
        elif x_rank == 3:
            pass
        else:
            raise ValueError("Cannot handle inputs of rank {}.".format(x_rank))

        # Apply layers one by one.
        for layer in self.layers:
            x = layer(x)

        # Remove batch dimension, if that wasn't specified initially.
        if x_rank != 3:
            x = x[0, :, :]

        return x

    def num_weights(self, input_size):
        num_weights = 0
        for layer in self.layers:
            num_weights += layer.num_weights(input_size)
            input_size = layer.width
        return num_weights

    @property
    def width(self):
        return self.layers[-1].width


class Recurrent(Layer):
    """A recurrent layer.

    Args:
        cell (:class:`.net.Cell`): Cell of the recurrent layer.
        h0 (tensor, optional): Initial hidden state. Defaults to a variable
            that will be learned.
    """

    def __init__(self, cell, h0=None):
        self.cell = cell
        self.h0 = h0

    def initialise(self, input_size, vs):
        if self.h0 is None:
            self.h0 = vs.get(shape=[1, self.cell.width])
        self.cell.initialise(input_size, vs)

    def __call__(self, x):
        # Put the batch dimension second.
        x_rank = B.rank(x)
        if x_rank == 2:
            x = x[:, None, :]
        elif x_rank == 3:
            x = B.transpose(x, perm=(1, 0, 2))
        else:
            raise ValueError(f"Cannot handle inputs of rank {B.rank(x)}.")

        # Recurrently apply the cell.
        n, batch_size, m = B.shape(x)
        y0 = B.zeros(B.dtype(x), batch_size, self.cell.width)
        h0 = B.tile(self.h0, batch_size, 1)
        res = B.scan(self.cell, x, h0, y0)[1]

        # Put the batch dimension first again.
        res = B.transpose(res, perm=(1, 0, 2))

        # Remove the batch dimension, if that didn't exist before.
        if x_rank == 2:
            res = res[0, :, :]

        return res

    def num_weights(self, input_size):
        return self.cell.width + self.cell.num_weights(input_size)

    @property
    def width(self):
        return self.cell.width


class Cell(metaclass=ABCMeta):
    """An abstract cell of a recurrent layer."""

    @abstractmethod
    def initialise(self, input_size, vs):
        """Initialise the weights of the cell.

        Args:
            input_size (int): Size of the input.
            vs (:class:`varz.Vars`): Variable container.
        """

    @abstractmethod
    def __call__(self, prev, x):  # pragma: no cover
        pass

    @abstractmethod
    def num_weights(self, input_size):
        """Get the number of weights used in the cell.

        Args:
            input_size (int): Size of the input.

        Returns:
            int: Number of weights.
        """

    @property
    @abstractmethod
    def width(self):
        """Width of the layer"""


class GRU(Cell):
    """A gated recurrent unit.

    Args:
        width (int): Width of the hidden.
        nonlinearity (function, optional): Nonlinearity to use. Defaults to
            `B.tanh`.
    """

    def __init__(self, width, nonlinearity=B.tanh):
        self._width = width
        self.f_z = Net(Linear(width), Activation(B.sigmoid))
        self.f_r = Net(Linear(width), Activation(B.sigmoid))
        self.f_h = Net(Linear(width), Activation(nonlinearity))

    def initialise(self, input_size, vs):
        self.f_z.initialise(self._width + input_size, vs)
        self.f_r.initialise(self._width + input_size, vs)
        self.f_h.initialise(self._width + input_size, vs)

    def __call__(self, prev, x):
        prev_h, _ = prev

        # Gate logic:
        z = self.f_z(B.concat(prev_h, x, axis=1))
        r = self.f_r(B.concat(prev_h, x, axis=1))
        h = (1 - z) * prev_h + z * self.f_h(B.concat(r * prev_h, x, axis=1))
        y = h

        return h, y

    def num_weights(self, input_size):
        return (
            self.f_z.num_weights(self._width + input_size)
            + self.f_r.num_weights(self._width + input_size)
            + self.f_h.num_weights(self._width + input_size)
        )

    @property
    def width(self):
        return self._width


class Elman(Cell):
    """Elman cell.

    Args:
        width (int): Width of the layer.
        nonlinearity (function, optional): Nonlinearity to use. Defaults to
            `B.tanh`.
    """

    def __init__(self, width, nonlinearity=B.tanh):
        self._width = width
        self.f_h = Net(Linear(width), Activation(nonlinearity))

    def initialise(self, input_size, vs):
        self.f_h.initialise(self._width + input_size, vs)

    def __call__(self, prev, x):
        prev_h, prev_y = prev

        # Gate logic:
        h = self.f_h(B.concat(prev_h, x, axis=1))
        y = h

        return h, y

    def num_weights(self, input_size):
        return self.f_h.num_weights(self._width + input_size)

    @property
    def width(self):
        return self._width


def ff(output_size, widths, nonlinearity=B.relu, normalise=True):
    """A standard feed-forward neural net.

    Args:
        output_size (int): Size of output.
        widths (tuple of int): Widths of the layers.
        nonlinearity (function, optional): Nonlinearity to use. Defaults to
            `B.relu`.
        normalise (bool, optional): Interleave with normalisation layers.
            Defaults to `True`.

    Returns:
        :class:`.net.Net`: Feed-forward NN.
    """
    layers = []

    for width in widths:
        layers.append(Linear(width))
        layers.append(Activation(nonlinearity))

        # Add a normalisation layer, if asked for.
        if normalise:
            layers.append(Normalise())

    # Add a final linear layer to scale outputs.
    layers.append(Linear(output_size))
    return Net(*layers)


def rnn(
    output_size,
    widths,
    nonlinearity=B.relu,
    normalise=False,
    gru=True,
    final_dense=False,
):
    """A standard recurrent neural net.

    Args:
        output_size (int): Size of output.
        widths (tuple of int): Widths of the layers.
        nonlinearity (function, optional): Nonlinearity to use. Defaults to
            `B.relu`.
        normalise (bool, optional): Interleave with normalisation layers.
            Defaults to `False`.
        gru (bool, optional): Use GRU layers instead of Elman layers. Defaults
            to `True`.
        final_dense (bool, optional): Append a final dense layer. Defaults to
            `False`.

    Returns:
        :class:`.net.Net`: RNN.
    """
    layer_type = GRU if gru else Elman
    layers = []

    for width in widths:
        layers.append(Recurrent(layer_type(width)))
        layers.append(Activation(nonlinearity))

        # Add a normalisation layer, if asked for.
        if normalise:
            layers.append(Normalise())

    # Add a final dense layer, if asked for.
    if final_dense:
        layers.append(Linear(output_size))
        layers.append(Activation(nonlinearity))

    # Add a final linear layer to scale outputs.
    layers.append(Linear(output_size))
    return Net(*layers)



import logging
import sys
import time
from collections import defaultdict

import lab as B
import numpy as np
from plum import Dispatcher

__all__ = ["Section", "out", "kv", "format", "Counter", "Progress"]

log = logging.getLogger(__name__)

_dispatch = Dispatcher()

indent_level = 0  #: Indentation level.
indent_width = 4  #: Number of indentation characters per level.
indent_char = " "  #: Indentation character.
key_width = 10  #: Minimum width for printing keys, for alignment.
streams = [sys.stdout]  #: Output streams.
digits = 3  #: Number of digits to show.
report_time = False  #: Report time.

_time_start = time.time()  #: Start time of the program.
_time_last_report = None  #: Last logged time stamp.


def _print(msg, line_end="\n", indent=True, flush=False):
    """Print a line with proper indentation.

    Args:
        msg (str): Message to print.
        line_end (str, optional): String to print at the end of the line.
            Defaults to "\n".
        indent (bool, optional): Insert indentation. Defaults to `True`.
        flush (bool, optional): Flush after writing. Defaults to `False`.
    """
    global indent_level
    global indent_width
    global indent_char

    global report_time
    global _time_start
    global _time_last_report

    if indent:
        # Construct total indentation.
        msg_indent = indent_char * indent_width * indent_level

        # Indent the message and all newlines therein.
        msg = msg_indent + msg.replace("\n", "\n" + msg_indent)

    if report_time and indent:
        time_delta = time.time() - _time_start

        # Calculate passed hours, minutes, and seconds.
        hours = int(time_delta / (60 * 60))
        time_delta -= 60 * 60 * hours
        minutes = int(time_delta / 60)
        time_delta -= 60 * minutes
        seconds = int(time_delta)

        # Build time report.
        time_report = f"{hours:02d}:{minutes:02d}:{seconds:02d}"

        # Only print the time report if it is different from last report. Else,
        # just indent.
        msg_indent = " " * len(time_report) + " | "
        if time_report != _time_last_report:
            msg = time_report + " | " + msg.replace("\n", "\n" + msg_indent)
            _time_last_report = time_report
        else:
            msg = msg_indent + msg.replace("\n", "\n" + msg_indent)

    # Write the message plus line end.
    for stream in streams:
        stream.write(msg + line_end)

        if flush:
            stream.flush()


class Section:
    """Create a section in which output is indented.

    Args:
        name (str): Name of the section.
    """

    def __init__(self, name=None):
        self.name = name

    def __enter__(self):
        # Print the section name, if one is given.
        if self.name is not None:
            _print("{}:".format(self.name))

        # Increase indentation.
        global indent_level
        indent_level += 1

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        # Decrease indentation.
        global indent_level
        indent_level -= 1


def out(msg):
    """Print a message.

    Args:
        msg (str): Message to print.
    """
    _print(msg)


@_dispatch
def kv(key, value, fmt=None, unit=None):
    """Print a key-value pair.

    Args:
        key (str): Key.
        value (object): Value.
        fmt (str, optional): Format for the value. Defaults to automatic
            formatting.
        unit (str, optional): Unit of the value. Defaults to no unit.
    """
    global key_width

    # Construct the format.
    f = "{{key:{width}}} {{value}}".format(width=key_width + 1)

    # Add the unit to the format, if it is given.
    if unit:
        f += " " + unit

    # Format the key.
    formatted_key = format(key)

    # Format the value.
    if fmt:
        # Use the given format.
        formatted_value = ("{:" + fmt + "}").format(value)
    else:
        # Automatically format.
        formatted_value = format(value)

    # If the value contains a newline, print it within a section.
    if "\n" in formatted_value:
        with Section(name=formatted_key):
            out(formatted_value)
    else:
        _print(f.format(key=formatted_key + ":", value=formatted_value))


@_dispatch
def kv(dict_: dict, **kw_args):
    for k, v in dict_.items():
        kv(k, v, **kw_args)


@_dispatch
def kv(key, dict_: dict, **kw_args):
    with Section(key):
        for k, v in dict_.items():
            kv(k, v, **kw_args)


@_dispatch
def format(x, info: bool):
    """Format an object.

    Args:
        x (object): Object to format.
        info (bool, optional): Print additional info, such as the shape of an
            object, where appropriate. Defaults to `True`.

    Returns:
        str: `x` as a string.
    """
    return str(x)


@_dispatch
def format(x):
    return format(x, True)


@_dispatch
def format(x: B.Number, info: bool):
    global digits

    # Format number in scientific notation.
    out = "{{:.{}e}}".format(digits).format(x)

    # If exponent is in [-2, 2], format as float instead.
    if "e" in out and -2 <= int(out.split("e")[1]) <= 2:
        return str(float(out))

    return out


@_dispatch
def format(x: B.Int, info: bool):
    return str(x)


@_dispatch
def format(xs: list, info: bool):
    return "[{}]".format(", ".join([format(x, info) for x in xs]))


@_dispatch
def format(xs: tuple, info: bool):
    return "({})".format(", ".join([format(x, info) for x in xs]))


@_dispatch
def format(xs: set, info: bool):
    return "{{{}}}".format(", ".join([format(x, info) for x in xs]))


@_dispatch
def format(x: B.NPNumeric, info: bool):
    # A NumPy array can be a scalar.
    if x.shape == ():
        return format.invoke(B.Number, bool)(x, info)

    # Represent as an array.
    x_str = np.array_str(x, precision=digits)

    # If additional information is requested and the array is displayed on
    # multiple lines, also show its shape and data type.
    if info and "\n" in x_str:
        x_str = (
            "({} array of data type {})\n"
            "".format("x".join([str(d) for d in x.shape]), x.dtype) + x_str
        )

    return x_str


@_dispatch
def format(x: B.Numeric, info: bool):
    return format(B.to_numpy(x), info)


class Counter:
    """A counter.

    Args:
        name (str, optional): Name of the counter. Defaults to "Counting".
        total (int, optional): Total number of counts. Defaults to no total.
    """

    def __init__(self, name="Counting", total=None):
        self.name = name
        self.total = total

    def __enter__(self):
        # Reset the counter.
        self.iteration = 0

        # Print the name.
        title = self.name

        # Print the total, if one is given.
        if self.total:
            title += " (total: {})".format(self.total)

        # Perform printing and flush, because there is no newline.
        _print(title + ":", line_end="", flush=True)

        return self

    def count(self):
        """Count once."""
        self.iteration += 1
        _print(" {}".format(self.iteration), line_end="", indent=False, flush=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        _print("\n", line_end="", indent=False, flush=True)

    @staticmethod
    def map(f, xs, name="Mapping"):
        """Perform a mapping operation that is counted.

        Args:
            f (function): Function to apply.
            xs (iterable): Arguments to apply.
            name (str, optional): Name of the mapping operation. Defaults to
                "Mapping".

        Returns:
            list: Result.
        """
        # Perform mapping operation.
        results = []
        with Counter(name=name, total=len(xs)) as counter:
            for x in xs:
                counter.count()
                results.append(f(x))
        return results


def _compute_alpha(cutoff_lag):
    """Compute the coefficient `a` of the one-pole filter
    `y[n] = a x[n] + (1 - a) x[n - 1]`.

    Args:
        cutoff_lag (int): Cut-off frequency, in number of lags.

    Returns:
        float: Coefficient `a`.
    """
    if cutoff_lag is None:
        return 1
    else:
        a = np.cos(2 * np.pi / cutoff_lag)
        return a - 1 + np.sqrt(a**2 - 4 * a + 3)


class Progress:
    """Display progress.

    Args:
        name (str): Name of operation. Defaults to "Progress".
        total (int): Total number of iterations. Defaults to no total.
        filter (dict): A dictionary mapping names of recorded values to the
            cut-off frequency of a one-pole filter, in number of lags.
        interval (int or float): Interval at which to print a report in
            either number of seconds (float) or number of iterations (int).
            Defaults to one second.
        filter_global (int): Cut-off frequency of a one-pole filter of the
            default smoother, in number of lags.

    """

    def __init__(
        self, name="Progress", total=None, filter=None, interval=1.0, filter_global=5
    ):
        self.name = name
        self.total = total
        self.interval = interval
        self.section = Section(name)

        global_alpha = _compute_alpha(filter_global)
        self.alpha = defaultdict(lambda: global_alpha)
        if filter:
            for name, cutoff in filter.items():
                self.alpha[name] = _compute_alpha(cutoff)

        self.cur_time = None
        self.iteration = None
        self.last_report = None
        self.report_was_just_shown = False
        self.values = dict()

    def __enter__(self):
        self.start_time = time.time()
        self.last_time = time.time()
        self.last_report = -np.inf
        self.report_was_just_shown = False
        self.iteration = 0
        self.values.clear()
        self.section.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if not self.report_was_just_shown:
            self.report()
        out("Done!")
        self.section.__exit__(exc_type, exc_val, exc_tb)

    @_dispatch
    def __call__(self, **updates):
        return self(updates)

    @_dispatch
    def __call__(self, updates: dict):
        now = time.time()

        # Increase iteration counter.
        self.iteration += 1

        # Record time per iteration.
        updates["_time_per_it"] = now - self.last_time
        self.last_time = now

        # Update tracked values.
        for name, value in updates.items():

            # Perform smoothing if `value` is numeric.
            if name in self.values and isinstance(value, B.Numeric):
                alpha = self.alpha[name]
                value = alpha * value + (1 - alpha) * self.values[name]

            self.values[name] = value

        # See if a progress report should be displayed.
        if (
            # If `self.interval` is a float, it specifies the number of
            # seconds between reports.
            (
                isinstance(self.interval, float)
                and now - self.last_report > self.interval
            )
            or
            # If `self.interval` is an int, it specifies the number of
            # iterations between reports.
            (isinstance(self.interval, int) and self.iteration % self.interval == 0)
            or
            # Always show a report on the first iteration.
            self.iteration == 1
        ):
            self.report()
        else:
            self.report_was_just_shown = False

    def report(self):
        """Show a report."""
        self.report_was_just_shown = True

        # Construct title.
        title = "Iteration {}".format(self.iteration)
        if self.total:
            title += "/" + str(self.total)

        with Section(title):
            # Show how much time has elapsed.
            kv("Time elapsed", "{:.1f} s".format(self.last_time - self.start_time))

            if self.total and "_time_per_it" in self.values:
                # Estimate time left.
                time_left = (
                    max(self.total + 1 - self.iteration, 0)
                    * self.values["_time_per_it"]
                )
                kv("Time left", "{:.1f} s".format(time_left))

            # Print all updates.
            for name in sorted(self.values.keys()):
                if name == "_time_per_it":
                    continue
                kv(name, self.values[name])

        # Record time of report.
        self.last_report = time.time()

    @staticmethod
    def map(f, xs, name="Mapping"):
        """Perform a mapping operation whose progress is shown.

        Args:
            f (function): Function to apply.
            xs (iterable): Arguments to apply.
            name (str, optional): Name of the mapping operation. Defaults to
                "Mapping".

        Returns:
            list: Result.
        """
        # Perform mapping operation.
        results = []
        with Progress(name=name, total=len(xs)) as progress:
            for x in xs:
                progress()
                results.append(f(x))
        return results



import abc

__all__ = ["Parser", "Literal", "Whitespace", "Float", "Integer", "SkipUntil"]


class Parser:
    """Parse a string.

    Args:
        path (str): Path to file parse.
    """

    def __init__(self, path):
        with open(path, "r") as f:
            self.lines = f.read().split("\n")

    def find_line(self, match):
        """Find the next line containing a string.

        Args:
            match (str): String to match.
        """
        while match.lower() not in self.lines[0].lower():
            self.next_line()

            if not self.lines_left():
                raise RuntimeError(
                    f'Attempting to match "{match.lower()}", but no lines left.'
                )

    def next_line(self):
        """Go to the next line."""
        self.lines = self.lines[1:]

    def lines_left(self):
        """Check if there are any lines left.

        Returns:
            bool: `True` if any lines are left, otherwise `False`.
        """
        return len(self.lines) > 0

    def parse(self, *parts):
        """Parse the current line.

        Args:
            *parts (tuple[:class:`.parser.RegexPart`]): Parts of the line.
        """
        res = Regex(*parts).parse(self.lines[0])
        return res[0] if len(res) == 1 else res


class Regex:
    """Regex.

    Args:
        *parts (tuple[:class:`.parser.RegexPart`]): Parts of the regex.
    """

    def __init__(self, *parts):
        self.parts = parts

    def parse(self, data):
        """Parse data according to the regex.

        Args:
            data (str): Data to parse.

        Returns:
            list: Matched results.
        """
        results = []
        for part in self.parts:
            data, results = part.parse(data, results)
        return results


class RegexPart(metaclass=abc.ABCMeta):
    """Part of a regex."""

    @abc.abstractmethod
    def parse(self, data, results):
        """Parse data and match results.

        Args:
            data (str): Data to parse.
            results (list): Matches so far.

        Returns:
            tuple[str, list]: Data left to parse and updated results.
        """


class Whitespace(RegexPart):
    """Whitespace."""

    def parse(self, data, results):
        while data[0] in {" ", "\n"}:
            data = data[1:]
        return data, results


class Float(RegexPart):
    """Floating point number."""

    def parse(self, data, results):
        num = ""
        while len(data) > 0 and data[0] in {  # Must be characters left to parse
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
            ".",
            "e",
            "-",
            "+",
            "N",
            "a",
        }:
            num += data[0]
            data = data[1:]
        return data, results + [float(num)]


class Integer(RegexPart):
    """Integer."""

    def parse(self, data, results):
        num = ""
        while len(data) > 0 and data[0] in {  # Must be characters left to parse
            "0",
            "1",
            "2",
            "3",
            "4",
            "5",
            "6",
            "7",
            "8",
            "9",
        }:
            num += data[0]
            data = data[1:]
        return data, results + [int(num)]


class Literal(RegexPart):
    """Literal.

    Args:
        literal (str): String to match.
    """

    def __init__(self, literal):
        self.literal = literal

    def parse(self, data, results):
        i = 0
        while i < len(self.literal):
            if data[0] == self.literal[i]:
                # Match! Go on.
                i += 1
                data = data[1:]
            else:
                raise RuntimeError(
                    f'When parsing literal "{self.literal}" at '
                    f"position {i + 1}, expected "
                    f'"{self.literal[i]}" but got "{data[0]}".'
                )
        return data, results


class SkipUntil(RegexPart):
    """Skip until a certain character is found. Then also skip that
    character.

    Args:
        char (str): Character to find.
    """

    def __init__(self, char):
        self.char = char

    def parse(self, data, results):
        self._check_data_left(data)
        while data[0] != self.char:
            data = data[1:]
            self._check_data_left(data)
        return data[1:], results  # Also skip the character.

    def _check_data_left(self, data):
        if len(data) == 0:
            raise RuntimeError(
                f"When attempting to skip until character "
                f'"{self.char}", the line was exhausted.'
            )
